/* -------------------------------------------------------------------*/
/*  GPS - a gps external based on libgps for pd                       */
/* -------------------------------------------------------------------*/
/* Written by Michael Aschauer <m@ash.to>                             */
/*                                                                    */
/* This library is free software; you can redistribute it and/or      */
/* modify it under the terms of the GNU Lesser General Public         */
/* License as published by the Free Software Foundation; either       */
/* version 2 of the License, or (at your option) any later version.   */
/*                                                                    */
/* This library is distributed in the hope that it will be useful,    */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  */
/* Lesser General Public License for more details.                    */
/*                                                                    */
/* You should have received a copy of the GNU Lesser General Public   */
/* License along with this library; if not, write to the              */
/* Free Software Foundation, Inc., 59 Temple Place - Suite 330,       */
/* Boston, MA  02111-1307, USA.                                       */
/* -------------------------------------------------------------------*/

#include "m_pd.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <gps.h>

#define EARTH_RADIUS_METER 6371.2 * 1000.0;

static t_class *gps_class;

typedef struct gps
{
    t_object x_obj;   
    
	t_outlet *lat_out, *lon_out, *alt_out, *spd_out;
	t_outlet *stat_out, *mode_out, *time_out, *track_out, *climb_out;
	t_outlet *utc_out;
	t_outlet *err_out;
	
	double last_time;
	double xx0,yy0,zz0;
	
	int verbose;
	int datemode;
	
	char *gpsd_host;
	char gpsd_port[8];
	
	struct gps_data_t *gpsdata;
	struct gps_fix_t *last_fix;
    
} t_gps;



void pd_gps_ouput_values(t_gps *x)
{	
	int i;
	t_atom err_list[7];
	char utc[128];
	char t_stamp[128];

	outlet_float(x->stat_out, (x->gpsdata->online > 0));		

	if(x->gpsdata->online > 0)
	{
		gps_poll(x->gpsdata);
		sprintf(t_stamp,"%f",x->gpsdata->online);
		
		switch(x->datemode)
		{
			case 1: 
				sprintf(t_stamp,"%f",x->gpsdata->online);
				sprintf(utc,"%f",x->gpsdata->fix.time);				
				break;
			case 0:
			default:
				unix_to_iso8601(x->gpsdata->fix.time,utc,20);
				unix_to_iso8601(x->gpsdata->online,t_stamp,20);
		}		
		outlet_symbol (x->time_out, gensym(t_stamp));
		
		//if(gps_query(x->gpsdata,"pvadt\n"))
		 
		if(x->gpsdata->fix.mode > MODE_NO_FIX && x->gpsdata->fix.time > x->last_time)
		{			
			
			outlet_float (x->mode_out, x->gpsdata->fix.mode);	
								
	
			outlet_float (x->lat_out, x->gpsdata->fix.latitude);
			outlet_float (x->lon_out, x->gpsdata->fix.longitude);		

			outlet_float (x->spd_out, x->gpsdata->fix.speed * KNOTS_TO_KPH);

			if(x->gpsdata->fix.track != -1)
				outlet_float (x->track_out, x->gpsdata->fix.track);
			
			if(x->gpsdata->fix.mode > MODE_2D)
			{
				if(x->gpsdata->fix.altitude != -999)
					outlet_float (x->alt_out, x->gpsdata->fix.altitude);							
				outlet_float (x->climb_out, x->gpsdata->fix.climb);	
			}

			outlet_symbol(x->utc_out, gensym(utc));
			
			// error estimation list
			SETFLOAT(err_list, x->gpsdata->fix.ept);
			SETFLOAT(err_list+1, x->gpsdata->fix.ept);
			SETFLOAT(err_list+2, x->gpsdata->fix.eph);
			SETFLOAT(err_list+3, x->gpsdata->fix.epv);
			SETFLOAT(err_list+4, x->gpsdata->fix.eps);
			SETFLOAT(err_list+5, x->gpsdata->fix.epc);
			outlet_list(x->err_out, gensym("list"),6,err_list);
			

		}
		else
		{
			outlet_float (x->mode_out, x->gpsdata->fix.mode);					
			post("GPS: No valid/new Fix.");
		}	
	}


	
	if(x->verbose)
	{
	
		post("GPS: online=%f",x->gpsdata->online);
		post("GPS: status=%d",x->gpsdata->status);	

		unix_to_iso8601(x->gpsdata->online,utc,20);
		post("GPS: time: %s",utc);
		post("GPS:-----------------------------------");
		post("GPS: FIX \ttimestamp: %f",x->gpsdata->fix.time);	
		post("GPS: mode:%d \tlat:%f \tlon:%f \talt:%f",
			x->gpsdata->fix.mode,
			x->gpsdata->fix.latitude,
			x->gpsdata->fix.longitude,
			x->gpsdata->fix.altitude);			

		post("GPS: \t\tspeed:%f \tclimb:%f \ttrack:%f",
			x->gpsdata->fix.speed * KNOTS_TO_KPH,
			x->gpsdata->fix.climb,
			x->gpsdata->fix.track);
			
		post("GPS: Error estimation:");
		post("GPS: \t\th=%f v=%f t=%f",
			x->gpsdata->fix.eph,
			x->gpsdata->fix.epv,			
			x->gpsdata->fix.ept);	
		post("GPS: \t\ttrck=%f spd=%f clmb=%f",
			x->gpsdata->fix.ept,
			x->gpsdata->fix.eps,			
			x->gpsdata->fix.epc);					
			
		post("GPS:-----------------------------------");		
		  		 			
		post("GPS: satellites used: %d - satellites in view: %d",
			x->gpsdata->satellites_used,
			x->gpsdata->satellites);	

		for (i=0;i<MAXCHANNELS;i++)
		{
			post("GPS: channel[%d] used=%d, elevation=%d, azimuth=%d, signal=%d",
				i, x->gpsdata->used[i], x->gpsdata->elevation[i], 
				x->gpsdata->azimuth[i], x->gpsdata->ss[i]);
		}
	}
	
	x->last_fix = &x->gpsdata->fix;	
}


void pd_gps_set(t_gps *x, t_floatarg f)
{
}

void pd_gps_setverbose(t_gps *x, t_floatarg f)
{
	x->verbose = f;
}

void pd_gps_setdatemode(t_gps *x, t_floatarg f)
{
	x->datemode = f;
}

void pd_gps_reset(t_gps *x)
{}


void pd_gps_bang(t_gps *x)
{
	pd_gps_ouput_values(x);
}


void pd_gps_connect(t_gps *x)
{
	if(x->gpsdata) gps_close(x->gpsdata);
	x->gpsdata = gps_open(x->gpsd_host, x->gpsd_port);
	
	if (!(x->gpsdata)) 
	{
		post("gps: no gpsd running or network error on %s %s",x->gpsd_host, x->gpsd_port);
	}
	else
	{
		post("gps: connected to gpsd on %s %s",x->gpsd_host, x->gpsd_port);
    	//gps_set_raw_hook(x->gpsdata, pd_gps_ouput_values(x));
    	gps_query(x->gpsdata, "w+x\n");
	}
}

void pd_gps_open(t_gps *x, t_symbol *host, t_float port)
{
	x->gpsd_host = host->s_name;
	sprintf(x->gpsd_port,"%d",(int)port);
	
	pd_gps_connect(x);	
}


void gps_free(t_gps *x)
{
	if(x->gpsdata) gps_close(x->gpsdata);
	post("gps: connection closed");
  	/*free(x->gpsdata);*/
}

void *gps_new(void)
{
    t_gps *x = (t_gps *)pd_new(gps_class);	

	x->stat_out = outlet_new(&x->x_obj, &s_float);
	x->time_out = outlet_new(&x->x_obj, &s_symbol);
		
	x->mode_out = outlet_new(&x->x_obj, &s_float);	
	       	
	x->lat_out = outlet_new(&x->x_obj, &s_float);
	x->lon_out = outlet_new(&x->x_obj, &s_float);
	x->alt_out = outlet_new(&x->x_obj, &s_float);
	
	x->spd_out = outlet_new(&x->x_obj, &s_float);
	x->track_out = outlet_new(&x->x_obj, &s_float);
	x->climb_out = outlet_new(&x->x_obj, &s_float);

	x->err_out = outlet_new(&x->x_obj, &s_list);
	x->utc_out = outlet_new(&x->x_obj, &s_symbol);


	x->gpsd_host = "localhost";
	sprintf(x->gpsd_port,"%d",2947);
	x->gpsdata = NULL;
	x->last_fix = NULL;
	x->last_time = 0;
	x->datemode = 0;	
	
	pd_gps_connect(x);
	
    return (x);
}


void gps_setup(void)
{    
    gps_class = class_new(gensym("gps"), (t_newmethod)gps_new, 
		(t_method)gps_free, sizeof(t_gps), 0,0);

    class_addbang(gps_class, pd_gps_bang);
	
    class_addmethod(gps_class, (t_method)pd_gps_reset, gensym("reset"), 0);
    class_addmethod(gps_class, (t_method)pd_gps_open, gensym("open"),A_DEFSYMBOL, A_DEFFLOAT, 0);  	
	class_addmethod(gps_class, (t_method)pd_gps_set, gensym("set"),A_DEFFLOAT, 0);
	class_addmethod(gps_class, (t_method)pd_gps_setverbose, gensym("verbose"),A_DEFFLOAT, 0);
	class_addmethod(gps_class, (t_method)pd_gps_setdatemode, gensym("datemode"),A_DEFFLOAT, 0);	
	
	class_sethelpsymbol(gps_class, gensym("help-gps"));
}
